﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Runtime.Serialization.Formatters.Binary;
using System.Threading;
using Classification.kNN;
using Classification.NN;
using Plossum.CommandLine;
using XmlClassifier.Algorithms;
using XmlClassifier.Algorithms.Classification;
using XmlClassifier.DataStructures;
using XmlClassifier.DataStructures.Patterns;
using XmlClassifier.Helpers;

namespace XmlClassifier
{
	public class Program
	{
		public static void Main(string[] args)
		{
			Options options = ParseParameters();

			if (!options.Help)
			{
				List<Document> trainingSet = Read(options.Train);
				List<Document> testSet = Read(options.Test);

				switch (options.Algorithm)
				{
					case Algorithm.KNN:
						KNN(options, trainingSet, testSet);
						break;

					case Algorithm.KNP:
					case Algorithm.NP:
						NP(options, trainingSet, testSet);
						break;

					case Algorithm.Matching:
						Matching(options, trainingSet, testSet);
						break;

					case Algorithm.Rules:
						Rules(options, trainingSet, testSet);
						break;

					default:
						break;
				}

				DebugConsole.WriteLine();

				Console.WriteLine("Accuracy: " + Evaluation.Accuracy(testSet, Evaluation.CostModel.Percent));

				DebugConsole.ReadLine();
			}
		}

		private static void Rules(Options options, List<Document> trainingSet, List<Document> testSet)
		{
			List<BasePattern> patterns = Train(options, trainingSet, testSet);

			DebugConsole.WriteLine("Testing");

			XmlClassifier.Algorithms.Classification.Rules.Classify(patterns, ref testSet);
		}

		private static void KNN(Options options, List<Document> trainingSet, List<Document> testSet)
		{
			DebugConsole.WriteLine("Testing");

			foreach (var item in testSet)
			{
				DebugConsole.SetCursorPosition(0, Console.CursorTop);
				DebugConsole.Write(string.Format("{0,3:0}%", (double)(testSet.IndexOf(item)) * 100 / testSet.Count));

				kNN<Document>.Classify(trainingSet, item, options.K, options.EquidistantVotes);
			}
		}

		private static void NP(Options options, List<Document> trainingSet, List<Document> testSet)
		{
			List<BasePattern> patterns = Train(options, trainingSet, testSet);

			if (patterns.Count > 0)
			{
				DebugConsole.WriteLine("Testing");

				BasePattern.WeightSupport = options.WeightSupport;
				BasePattern.WeightConfidence = options.WeightConfidence;
				BasePattern.WeightDistance = options.WeightDistance;
				BasePattern.WeightSize = options.WeightSize;

				switch (options.Algorithm)
				{
					case Algorithm.KNP:
						foreach (var item in testSet)
						{
							DebugConsole.SetCursorPosition(0, Console.CursorTop);
							DebugConsole.Write(string.Format("{0,3:0}%", (double)(testSet.IndexOf(item)) * 100 / testSet.Count));

							kNP<BasePattern, Document>.Classify(patterns, item, options.K, options.EquidistantVotes);
						}
						break;

					case Algorithm.NP:
						foreach (var item in testSet)
						{
							DebugConsole.SetCursorPosition(0, Console.CursorTop);
							DebugConsole.Write(string.Format("{0,3:0}%", (double)(testSet.IndexOf(item)) * 100 / testSet.Count));

							NP<BasePattern, Document>.Classify(patterns, item, options.Range);
						}
						break;

					default:
						break;
				}
			}
		}

		private static void Matching(Options options, List<Document> trainingSet, List<Document> testSet)
		{
			List<BasePattern> patterns = Train(options, trainingSet, testSet);

			DebugConsole.WriteLine("Testing");

			Simple.Classify(patterns, ref testSet);
		}

		private static List<BasePattern> Train(Options options, List<Document> trainingSet, List<Document> testSet)
		{
			DebugConsole.WriteLine("Training");

			List<BasePattern> patterns = null;

			switch (options.Pattern)
			{
				case Pattern.Tree:
					patterns = GetTreePatterns(trainingSet, options);
					TreePattern.RemoveDuplicatePatterns(patterns, options.DuplicateTreatment);
					break;

				case Pattern.Path:
					PathPattern.TransformDocuments(trainingSet);
					PathPattern.TransformDocuments(testSet);

					patterns = GetPathPatterns(trainingSet, options);
					break;

				default:
					throw new NotImplementedException("Available pattern definitions are: Tree, Path.");
			}

			if (patterns.Count == 0)
			{
				throw new Exception("No Patterns found...");
			}
			else
			{
				foreach (var c in trainingSet.Select(x => x.ExpectedClass).Distinct())
				{
					DebugConsole.WriteLine("Klasa " + c + ": " + patterns.Where(x => x.Class == c).Count() + " wzorców");
					foreach (var p in patterns.Where(x => x.Class == c))
					{
						//DebugConsole.WriteLine(((TreePattern)p).Pattern);
					}
				}
			}

			return patterns;//.Where(x => x.Size > patterns.Min(y => y.Size)).ToList();
		}

		private static List<BasePattern> GetPathPatterns(List<Document> trainingSet, Options options)
		{
			PathPattern.MaximalPathsOnly = options.MaxPaths;
			PathPattern.MaximalPathLength = options.MaxPathLength;

			return (new PathPattern()).GetPatterns(trainingSet, options.MinSup);
		}

		private static List<BasePattern> GetTreePatterns(List<Document> trainingSet, Options options)
		{
			List<BasePattern> patterns = null;

			DirectoryInfo di = Directory.Exists("Patterns") ? new DirectoryInfo("Patterns") : Directory.CreateDirectory("Patterns");

			string filename = options.Train;
			if (filename.Contains("\\"))
			{
				filename = filename.Substring(filename.LastIndexOf("\\") + 1, filename.Length - filename.LastIndexOf("\\") - 1);
			}
			var files = di.GetFiles(filename + "_" + options.MinSup + ".txt");
			if (files.Length == 1 && !options.NoCache)
			{
				using (Stream stream = File.Open("Patterns\\" + filename + "_" + options.MinSup + ".txt", FileMode.Open))
				{
					BinaryFormatter bin = new BinaryFormatter();
					patterns = (List<BasePattern>)bin.Deserialize(stream);
				}
			}
			else
			{
				patterns = (new TreePattern()).GetPatterns(trainingSet, options.MinSup);

				if (!options.NoCache)
				{
					using (Stream stream = File.Open("Patterns\\" + filename + "_" + options.MinSup + ".txt", FileMode.Create))
					{
						BinaryFormatter bin = new BinaryFormatter();
						bin.Serialize(stream, patterns);
					}
				}
			}

			return patterns;
		}

		private static Options ParseParameters()
		{
			Thread.CurrentThread.CurrentCulture = new CultureInfo("en-US");

			Options options = new Options();
			CommandLineParser parser = new CommandLineParser(options, NumberFormatInfo.CurrentInfo);
			parser.Parse();
			if (!options.NoInfo)
			{
				Console.WriteLine(parser.UsageInfo.GetHeaderAsString(78));
			}

			if (options.Help)
			{
				Console.WriteLine(parser.UsageInfo.GetOptionsAsString(78));
			}
			else if (parser.HasErrors)
			{
				Console.WriteLine(parser.UsageInfo.GetErrorsAsString(78));
			}
			return options;
		}

		private static List<Document> Read(string path)
		{
			List<Document> result = new List<Document>();

			using (StreamReader sr = new StreamReader(path))
			{
				while (!sr.EndOfStream)
				{
					result.Add(new Document(sr.ReadLine()));
				}
			}

			return result;
		}
	}
}
